/* ************************************************************************
 * Copyright (C) 2025 Advanced Micro Devices, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * THE SOFTWARE IS PROVIDED "AS IS") WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * ************************************************************************ */
#pragma once

#include <cstdint>
#include <functional>
#include <unordered_map>
#include <vector>

#include "stinkytofu/Config/Config.h"

/* Begin architecture-specific ArchInfo headers */

// GFX1250
#ifdef STINKYTOFU_ARCH_GFX1250
#include "Gfx1250RocisaArchInfo.hpp"
#endif

/* End of architecture-specific ArchInfo headers */

namespace {
using RocisaToHwInstMap = std::unordered_map<std::type_index, uint16_t>;
using RocisaToHwInstLoweringMap =
    std::unordered_map<std::type_index, stinkytofu::ConvertRocisaToHwInstFunc>;

std::vector<std::function<const RocisaToHwInstMap*()>> vecRocisaToHwInstMap = {
#define STINKYTOFU_ARCH(archName) archName##RocisaToHwInstMap,
#include "Config/Archs.def"
};

std::vector<std::function<const RocisaToHwInstLoweringMap*()>> vecRocisaToHwInstLoweringMap = {
#define STINKYTOFU_ARCH(archName) archName##RocisaToHwInstLoweringMap,
#include "Config/Archs.def"
};
}  // namespace
